@@ -1,11 +1,13 @@
 from .api import connect
 from .objdb import ObjDBMixin
 from .dbobj import DbObj, stored
+from .base import BaseDB
 
 __all__ = [
     'connect',
     'stored',
     'ObjDBMixin',
     'DbObj',
-    'stored'
+    'stored',
+    'BaseDB'
 ]
